package com.threed.jpct.demos.terrain;

import java.awt.*;
import java.awt.event.*;
import com.threed.jpct.*;
import com.threed.jpct.util.*;
import com.threed.jpct.threading.*;

import java.io.*;
import java.util.*;
import java.util.List;

import org.lwjgl.input.*;

public class NewTerrain implements IPaintListener {

	private static final long serialVersionUID = 1L;

	private KeyMapper keyMapper = null;

	private FrameBuffer fb = null;
 
	private World world = null;

	private Object3D terrain = null;
	private Object3D terrain2 = null;
	private Object3D water = null;
	private Object3D water2 = null;

	private World sky = null;
	private Object3D dome = null;
	private Object3D sunObj = null;

	private int fps = 0;

	private long time = System.currentTimeMillis();

	private MouseMapper mouse = null;

	private float xAngle = 0;

	private boolean doloop = true;

	private boolean forward = false;
	private boolean backward = false;
	private boolean up = false;
	private boolean down = false;
	private boolean left = false;
	private boolean right = false;
	private boolean sunDown = false;
	private boolean sunUp = false;
	private boolean showTrees = true;
	private boolean showDetails = false;
	private boolean showReflections=true;
	private boolean showReflectionMap=false;

	private Light sun = null;

	private LensFlare lens = null;
	private Ticker ticker = new Ticker(20);

	private Color skyColor = new Color(195, 210, 230);

	private SimpleVector sunDir = new SimpleVector();

	private int treeCount = 1500;
	private List<Object3D> trees = new ArrayList<Object3D>(treeCount);
	private List<SimpleVector> treePos = new Vector<SimpleVector>(treeCount);

	private float ang = 0;
	private Random rnd = new Random(9);

	private boolean wireFrame = false;

	private int its;

	private int refMapSize=2048;

	/*	private int cnt = 0;
*/
	public NewTerrain() {

		Config.glAvoidTextureCopies = true;
		Config.maxPolysVisible = 5000;
		Config.glColorDepth = 24;
		Config.useMultipleThreads = Runtime.getRuntime().availableProcessors()>1;
		Config.glVertexArrays = true;
		Config.glFullscreen = false;
		Config.glShadowZBias = 0.0f;
		Config.glTrilinear = true;
		Config.glTransparencyOffset = 0.0f;
		Config.glTransparencyMul = 0.025f;
		Config.glUseFBO = true;
		Config.glIgnoreNearPlane = false;
	}

	public void finishedPainting() {
		fps++;
		if (System.currentTimeMillis() - time >= 1000) {
			System.out.println(fps);
			//System.out.println("Its: "+its);
			fps = 0;
			its=0;
			time = System.currentTimeMillis();
		}
	}

	public void startPainting() {
	}

	private void initStuff() throws Exception {
		fb = new FrameBuffer(1024, 768, FrameBuffer.SAMPLINGMODE_NORMAL);
		Config.glFullscreen = false;

		world = new World();
		sky = new World();

		sky.setClippingPlanes(1, 750000);
		world.setClippingPlanes(1, 20000);

		fb.enableRenderer(IRenderer.RENDERER_OPENGL, IRenderer.MODE_OPENGL);

		fb.disableRenderer(IRenderer.RENDERER_SOFTWARE);
		keyMapper = new KeyMapper();
		fb.setPaintListener(this);

		TextureManager tm = TextureManager.getInstance();
		Texture ground = new Texture("terrain/grass.jpg");
		tm.addTexture("ground", ground);
		Texture rocks = new Texture("terrain/rocks.jpg");
		tm.addTexture("rocks", rocks);
		Texture sand = new Texture("terrain/sand.jpg");
		tm.addTexture("sand", sand);
		Texture details = new Texture("terrain/details.jpg"); 
		tm.addTexture("details", details);

		Texture wat = new Texture("terrain/water2.jpg");
		tm.addTexture("water", wat);
		Texture wat2 = new Texture("terrain/water.jpg");
		tm.addTexture("water_blue", wat2);
		Texture suny = new Texture("terrain/sun.jpg");
		tm.addTexture("sun", suny);
		Texture skyy = new Texture("terrain/sky.jpg");
		tm.addTexture("sky", skyy);

		Texture treet = new Texture("terrain/tree.jpg");
		tm.addTexture("tree", treet);
		Texture treet2 = new Texture("terrain/tree2.jpg");
		tm.addTexture("tree2", treet2);

		details.setEnabled(false);
		
		terrain = Object3D.mergeAll(Loader.loadOBJ("terrain/terrain.obj", "terrain/terrain.mtl", 5f));
		terrain.rotateX(-(float) Math.PI / 2f);
		terrain.rotateMesh();
		terrain.setRotationMatrix(new Matrix());

		float[] bb = terrain.getMesh().getBoundingBox();
		float minX = bb[0];
		float maxX = bb[1];
		float minZ = bb[4];
		float maxZ = bb[5];
		float minY = bb[2];
		float maxY = bb[3];
		float dx = maxX - minX;
		float dz = maxZ - minZ;
		dx /= 20f;
		dz /= 20f;

		int tid = tm.getTextureID("ground");
		int sid = tm.getTextureID("sand");
		int did = tm.getTextureID("details");

		PolygonManager pm = terrain.getPolygonManager();
		for (int i = 0; i < pm.getMaxPolygonID(); i++) {
			SimpleVector v0 = pm.getTransformedVertex(i, 0);
			SimpleVector v1 = pm.getTransformedVertex(i, 1);
			SimpleVector v2 = pm.getTransformedVertex(i, 2);
			TextureInfo ti = new TextureInfo(tid, v0.x / dx, v0.z / dz, v1.x / dx, v1.z / dz, v2.x / dx, v2.z / dz);
			float dxd = dx / 10;
			float dzd = dz / 10;
			ti.add(did, v0.x / dxd, v0.z / dzd, v1.x / dxd, v1.z / dzd, v2.x/ dxd, v2.z / dzd, TextureInfo.MODE_BLEND);
			//ti.add(did, v0.x / dxd, v0.z / dzd, v1.x / dxd, v1.z / dzd, v2.x / dxd, v2.z / dzd, TextureInfo.MODE_MODULATE);
			pm.setPolygonTexture(i, ti);
		}

		terrain2 = new Object3D(terrain, true);
		terrain2.setSortOffset(-1000);
		terrain2.setTexture("rocks");
		world.addObject(terrain2);

		pm = terrain2.getPolygonManager();
		for (int i = 0; i < pm.getMaxPolygonID(); i++) {
			for (int p = 0; p < 3; p++) {
				SimpleVector v = pm.getTransformedVertex(i, p);
				if (v.y < -1200) {
					pm.setVertexAlpha(i, p, (float) Math.sqrt(2 * (v.y + 1200) / (minY + 1200)));
				} else {
					pm.setVertexAlpha(i, p, 0);
				}
				if (v.y > -200) {
					pm.setPolygonTexture(i, sid);
					pm.setVertexAlpha(i, p, (float) Math.sqrt((v.y + 200) / (maxY + 200)));
				}
			}
		}

		terrain2.setTransparency(0);
		terrain.compileAndStrip();
		terrain2.compileAndStrip();

		world.addObject(terrain);

		water = Primitives.getPlane(20, 500);
		water.scale(3.8f);
		water.rotateX((float) Math.PI / 2f);
		world.addObject(water);
		water.setTexture("water");
		water.compileAndStrip();

		world.setFoggingMode(World.FOGGING_PER_PIXEL);
		world.setFogging(World.FOGGING_ENABLED);
		world.setFogParameters(15000, 20000, skyColor.getRed(), skyColor.getGreen(), skyColor.getBlue());

		Config.lightMul = 1;

		Texture burst = new Texture("terrain/lens/lens1.jpg", true);
		Texture halo1 = new Texture("terrain/lens/lens2.jpg", true);
		Texture halo2 = new Texture("terrain/lens/lens3.jpg", true);
		Texture halo3 = new Texture("terrain/lens/lens4.jpg", true);

		tm.addTexture("burst", burst);
		tm.addTexture("halo1", halo1);
		tm.addTexture("halo2", halo2);
		tm.addTexture("halo3", halo3);

		sun = new Light(world);

		lens = new LensFlare(SimpleVector.ORIGIN, "burst", "halo1", "halo2", "halo3");
		lens.setTransparency(12);

		world.setAmbientLight(90, 90, 90);
		world.buildAllObjects();
		water.translate(terrain.getCenter());
		water.translate(0, 1500, 0);
		water.setTransparency(15);
		water.setSortOffset(-200000);
		reTexture(water, 40, true);
		
		System.out.println(water.getTranslation());

		water2 = new Object3D(water, true);
		world.addObject(water2);
		water2.shareCompiledData(water);
		water2.compileAndStrip();
		water2.translate(0, 10, 0);
		water2.setSortOffset(-190000);
		water2.setTexture("water_blue");
		reTexture(water2, 40, false);


		sunObj = Primitives.getPlane(1, 4000);
		sunObj.setTexture("sun");
		sunObj.setTransparency(12);
		sunObj.setTransparencyMode(Object3D.TRANSPARENCY_MODE_ADD);
		sunObj.setLighting(Object3D.LIGHTING_NO_LIGHTS);
		sunObj.setAdditionalColor(Color.WHITE);
		sunObj.setBillboarding(true);

		sky.addObject(sunObj);

		dome = Object3D.mergeAll(Loader.load3DS("terrain/dome.3ds", 5000));
		dome.setTexture("sky");
		sky.addObject(dome);
		dome.compileAndStrip();
		dome.setLighting(Object3D.LIGHTING_NO_LIGHTS);
		dome.setAdditionalColor(Color.WHITE);
		dome.rotateX(-(float) Math.PI / 2f);
		sky.buildAllObjects();
		dome.setTransparency(-1);

		Logger.log("Loading trees...", Logger.MESSAGE);
		Object3D tree = Object3D.mergeAll(Loader.load3DS("terrain/tree.3ds", 5));
		Object3D tree2 = Object3D.mergeAll(Loader.load3DS("terrain/tree2.3ds", 5));

		tree.rotateX((float) Math.PI);
		tree2.rotateX((float) Math.PI);
		tree.rotateMesh();
		tree2.rotateMesh();
		tree.setRotationMatrix(new Matrix());
		tree2.setRotationMatrix(new Matrix());

		tree.setTexture("tree");
		tree2.setTexture("tree2");
		tree.setTransparency(40);
		tree2.setTransparency(40);

		terrain.enableLazyTransformations();
		terrain2.enableLazyTransformations();

		for (int i = 0; i < treeCount; i++) {
			Object3D t = null;
			if (i % 10 == 0 && false) {
				t = new Object3D(tree2, true);

			} else {
				t = new Object3D(tree, true);
				if (i != 0) {
					t.shareCompiledData(trees.get(0));
				}
			}
			t.compileAndStrip();
			t.setVisibility(false);
			world.addObject(t);
			t.setSortOffset(-250000);
			trees.add(t);
		}
		populateTerrain(minX, maxX, minZ, maxZ);
	}

	private void populateTerrain(final float minX, final float maxX, final float minZ, final float maxZ) {
		Logger.log("Planting trees...", Logger.MESSAGE);
		terrain.setCollisionMode(Object3D.COLLISION_CHECK_OTHERS);
		OcTree oc = new OcTree(terrain.getMesh(), 500, OcTree.MODE_OPTIMIZED);
		terrain.setOcTree(oc);
		oc.setCollisionUse(true);
		oc.setRenderingUse(false);
		World.setDefaultThread(Thread.currentThread());

		Worker worker = new Worker(3);
		for (int i = 0; i < 3; i++) {
			worker.add(new WorkLoad() {
				public void doWork() {
					SimpleVector wl = water.getTransformedCenter();
					for (int i = 0; i < 5000;) {
						float nx, nz;
						synchronized (terrain) {
							nx = minX + (rnd.nextFloat() * (maxX - minX));
							nz = minZ + (rnd.nextFloat() * (maxZ - minZ));
						}
						float ny = -1400;

						float f = terrain.calcMinDistance(new SimpleVector(nx, ny, nz), new SimpleVector(0, 1, 0), 2000);

						if (f != Object3D.COLLISION_NONE) {
							SimpleVector s = new SimpleVector(nx, ny + f + 6, nz);

							if (s.y < wl.y - 100) {
								treePos.add(s);
								i++;
								if (i % 100 == 0) {
									Logger.log("Trees planted: " + treePos.size(), Logger.MESSAGE);
								}
							}
						}
					}
				}

				public void done() {
					// TODO Auto-generated method stub
				}

				public void error(Exception arg0) {
					// TODO Auto-generated method stub
				}
			});
		}
		worker.waitForAll();
		worker.dispose();
		terrain.setCollisionMode(Object3D.COLLISION_CHECK_NONE);
	}

	private void processTrees() {
		// This is slow and culls too much at the border of the viewport.
		// It's not meant to be used in a productive application.

		int tp = 0;

		int end = treePos.size();
		int endt = trees.size();

		if (showTrees) {
			SimpleVector pos = world.getCamera().getPosition();
			SimpleVector dir = world.getCamera().getDirection();

			for (int i = 0; i < end; i++) {
				SimpleVector pp = treePos.get(i);
				SimpleVector toTree = pp.calcSub(pos);
				float len = toTree.length();
				if (len < 10000 && Math.abs(toTree.calcAngle(dir)) < Math.PI / 3.5f) {
					Object3D t = trees.get(tp);
					t.setVisibility(true);
					t.getTranslationMatrix().setIdentity();
					t.translate(pp);
					t.getRotationMatrix().setIdentity();
					t.rotateY(pp.x);
					if (len > 6000) {
						t.setTransparency((4000 - ((int) len - 6000)) / 100);
					} else {
						t.setTransparency(40);
					}
					tp++;
					if (tp >= endt) {
						break;
					}
				}
			}
		}

		for (int i = tp; i < endt; i++) {
			Object3D tree = trees.get(i);
			tree.setVisibility(false);
		}
	}

	private void checkKeys(long ticks) {
		KeyState ks = null;
		while ((ks = keyMapper.poll()) != KeyState.NONE) {
			if (ks.getKeyCode() == KeyEvent.VK_ESCAPE) {
				doloop = false;
			}

			if (ks.getKeyCode() == KeyEvent.VK_UP) {
				forward = ks.getState();
			}

			if (ks.getKeyCode() == KeyEvent.VK_DOWN) {
				backward = ks.getState();
			}

			if (ks.getKeyCode() == KeyEvent.VK_LEFT) {
				left = ks.getState();
			}

			if (ks.getKeyCode() == KeyEvent.VK_RIGHT) {
				right = ks.getState();
			}

			if (ks.getKeyCode() == KeyEvent.VK_PAGE_UP) {
				up = ks.getState();
			}

			if (ks.getKeyCode() == KeyEvent.VK_PAGE_DOWN) {
				down = ks.getState();
			}
			if (ks.getKeyCode() == KeyEvent.VK_2) {
				sunUp = ks.getState();
			}
			if (ks.getKeyCode() == KeyEvent.VK_1) {
				sunDown = ks.getState();
			}
			if (ks.getKeyCode() == KeyEvent.VK_R && ks.getState()) {
				showReflections^= true;
			}
			if (ks.getKeyCode() == KeyEvent.VK_M && ks.getState()) {
				showReflectionMap^=true;
			}
			
			if (ks.getKeyCode() == KeyEvent.VK_T && ks.getState()) {
				showTrees^=true;
				processTrees();
			}

			if (ks.getKeyCode() == KeyEvent.VK_W && ks.getState()) {
				wireFrame^=true;
			}
			if (ks.getKeyCode() == KeyEvent.VK_D && ks.getState()) {
				showDetails^=true;
				TextureManager.getInstance().getTexture("details").setEnabled(showDetails);
			}
		}
		move(ticks);
	}

	private void move(long ticks) {

		if (ticks == 0) {
			return;
		}

		if (forward) {
			world.getCamera().moveCamera(Camera.CAMERA_MOVEIN, 50f * ticks);
		}

		if (backward) {
			world.getCamera().moveCamera(Camera.CAMERA_MOVEOUT, 50f * ticks);
		}

		if (left) {
			world.getCamera().moveCamera(Camera.CAMERA_MOVELEFT, 50f * ticks);
		}

		if (right) {
			world.getCamera().moveCamera(Camera.CAMERA_MOVERIGHT, 50f * ticks);
		}

		if (up) {
			world.getCamera().moveCamera(Camera.CAMERA_MOVEUP, 50f * ticks);
		}

		if (down) {
			world.getCamera().moveCamera(Camera.CAMERA_MOVEDOWN, 50f * ticks);
		}

		if (sunUp) {
			sunDir.rotateZ(ticks * 0.01f);
		}

		if (sunDown) {
			sunDir.rotateZ(ticks * -0.01f);
		}

		if (mouse == null) {
			mouse = new MouseMapper(fb);
			mouse.hide();
		}

		Matrix rot = world.getCamera().getBack();
		int dx = mouse.getDeltaX();
		int dy = mouse.getDeltaY();

		float ts = 0.2f;
		float tsy = ts;

		if (dx != 0) {
			ts = Math.abs(dx) / -500f;
		}
		if (dy != 0) {
			tsy = Math.abs(dy) / 500f;
		}

		if (dx < 0) {
			rot.rotateAxis(rot.getYAxis(), ts);
		}

		if (dx > 0) {
			rot.rotateAxis(rot.getYAxis(), -ts);
		}

		if (dy > 0 && xAngle < Math.PI / 4.2) {
			rot.rotateX(tsy);
			xAngle += tsy;
		}
		if (dy < 0 && xAngle > -Math.PI / 4.2) {
			rot.rotateX(-tsy);
			xAngle -= tsy;
		}
	}

	private void doIt() throws Exception {

		Camera cam = world.getCamera();
		cam.moveCamera(Camera.CAMERA_MOVEUP, 2000);
		cam.lookAt(terrain.getTransformedCenter());

		cam.setFOV(1.5f);
		
		sunDir = new SimpleVector(0, -20000, 0);
		SimpleVector sunPos = terrain.getTransformedCenter();

		sun.setIntensity(150, 150, 150);
		sun.setAttenuation(-1);

		world.getLights().setRGBScale(2);
		float a = 0;

		lens.setGlobalScale(3);
		Camera camy = sky.getCamera();

		SimpleVector up = new SimpleVector(0, -1, 0);
		float oldAng = 999;

		cam.setPosition(17552.824f, -1255.2219f, 9189.418f);

		sunPos.z += 20000;
		cam.rotateY(0.75f);
		cam.rotateX(0.27f);

		SimpleVector oldCam = new SimpleVector();
		SimpleVector oldPos = new SimpleVector();
		
		ReflectionHelper refHelper=new ReflectionHelper(fb, water, refMapSize, TextureInfo.MODE_MODULATE);
		refHelper.setClippingPlaneOffset(20);
		refHelper.exclude(water2);
		refHelper.setClearColor(skyColor);
		
		while (doloop && !org.lwjgl.opengl.Display.isCloseRequested()) {
			
			long ticks = ticker.getTicks();
			
			SimpleVector sunp = new SimpleVector(sunPos);
			sunp.add(sunDir);
			sun.setPosition(sunp);
			sunObj.getTranslationMatrix().setIdentity();
			sunObj.translate(sunp);
			lens.setLightPosition(sunp);

			sunp = sunp.calcSub(sunPos);
			SimpleVector sn = sunp.normalize();
			ang = sn.calcAngle(up);

			if (ang != oldAng) {
				oldAng = ang;
				float cos = (float) Math.cos(ang);
				world.setAmbientLight((int) (90f * Math.sqrt(cos) + 10), (int) (90f * cos + 10), (int) (90f * cos + 10));
				sun.setIntensity((int) (140f * Math.sqrt(cos) + 10), (int) (140f * cos + 10), (int) (140f * cos + 10));
				if (cos > 0) {
					Color col = new Color((int) (245f * cos + 10), (int) (245f * cos * cos + 10), (int) (245f * cos * cos + 10));
					skyColor = new Color((int) (195f * Math.sqrt(Math.sqrt(cos))), (int) (210f * Math.sqrt(cos)), (int) (230f * Math.sqrt(cos)));
					world.setFogParameters(15000, 20000, skyColor.getRed(), skyColor.getGreen(), skyColor.getBlue());
					dome.setAdditionalColor(col);
					sunObj.setAdditionalColor(col);
				}
			}

			checkKeys(ticks);
			camy.setPosition(cam.getPosition());
			camy.getBack().setTo(cam.getBack());

			if (ticks > 0) {
				water.translate((float) Math.sin(a), (float) Math.sin(a) / 20f, (float) Math.cos(a));
				water2.translate((float) -Math.sin(a), (float) Math.cos(a) / 20f, -(float) Math.cos(a));
				a += ticks * 0.01f;
			}
			
			if (ticks > 0 && (!oldCam.equals(cam.getDirection()) || !oldPos.equals(cam.getPosition()))) {
				processTrees();
				oldCam.set(cam.getDirection());
				oldPos.set(cam.getPosition());
			}
		
			if (showReflections) {
				refHelper.setEnabled(true);
				refHelper.drawScene(world, fb, new World[]{sky});
			} else {
				refHelper.setEnabled(false);
			}
			
			// Render to screen
			fb.clear(skyColor);
			fb.setPaintListenerState(false);
			sky.renderScene(fb);
			sky.draw(fb);
			fb.setPaintListenerState(true);
			
			world.renderScene(fb);
			world.draw(fb);

			if (wireFrame) {
				fb.setPaintListenerState(false);
				world.drawWireframe(fb, Color.RED);
				fb.setPaintListenerState(true);
			}

			lens.update(fb, world);
			lens.render(fb);

			if (showReflectionMap) {
				fb.blit(refHelper.getReflectionTexture(), 0, 0, 0, 0, refMapSize, refMapSize, 256, 256, -1, false, null);
			}
			
			fb.update();
			fb.displayGLOnly();

			dome.rotateY(0.001f * ticks);
			its++;
		}
		fb.dispose();
		System.exit(0);
	}

	public static void main(String[] args) throws Exception {
		Config.glVerbose = true;
		NewTerrain cd = new NewTerrain();
		cd.initStuff();
		cd.doIt();
	}

	public class SimpleStream {
		private InputStream stream;

		public SimpleStream(String file) {
			stream = this.getClass().getClassLoader().getResourceAsStream(file);
		}

		public InputStream getStream() {
			return stream;
		}

		public void close() throws IOException {
			stream.close();
		}
	}

	private void reTexture(Object3D obj, float tileFactor, boolean withReflection) {
		Logger.log("ReTexturing...", Logger.MESSAGE);
		PolygonManager pm = obj.getPolygonManager();
		int end = pm.getMaxPolygonID();
		int refid=TextureManager.getInstance().getTextureID("reflect");
		for (int i = 0; i < end; i++) {
			int t1 = pm.getPolygonTexture(i);
			SimpleVector uv0 = pm.getTextureUV(i, 0);
			SimpleVector uv1 = pm.getTextureUV(i, 1);
			SimpleVector uv2 = pm.getTextureUV(i, 2);
			uv0.scalarMul(tileFactor);
			uv1.scalarMul(tileFactor);
			uv2.scalarMul(tileFactor);

			TextureInfo ti=null;
			if (withReflection) {
				ti = new TextureInfo(t1, uv0.x, uv0.y, uv1.x, uv1.y, uv2.x, uv2.y);
				ti.add(refid, TextureInfo.MODE_MODULATE);
			} else {
				ti = new TextureInfo(t1, uv0.x, uv0.y, uv1.x, uv1.y, uv2.x, uv2.y);
			}

			pm.setPolygonTexture(i, ti);
		}
	}

	class MouseMapper {

		private boolean hidden = false;

		private int height = 0;

		public MouseMapper(FrameBuffer buffer) {
			height = buffer.getOutputHeight();
			init();
		}

		public void hide() {
			if (!hidden) {
				Mouse.setGrabbed(true);
				hidden = true;
			}
		}

		public void show() {
			if (hidden) {
				Mouse.setGrabbed(false);
				hidden = false;
			}
		}

		public boolean isVisible() {
			return !hidden;
		}

		public void destroy() {
			show();
			if (Mouse.isCreated()) {
				Mouse.destroy();
			}
		}

		public boolean buttonDown(int button) {
			return Mouse.isButtonDown(button);
		}

		public int getMouseX() {
			return Mouse.getX();
		}

		public int getMouseY() {
			return height - Mouse.getY();
		}

		public int getDeltaX() {
			if (Mouse.isGrabbed()) {
				return Mouse.getDX();
			} else {
				return 0;
			}
		}

		public int getDeltaY() {
			if (Mouse.isGrabbed()) {
				return Mouse.getDY();
			} else {
				return 0;
			}
		}

		private void init() {
			try {
				if (!Mouse.isCreated()) {
					Mouse.create();
				}

			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}
	}

	private static class Ticker {

		private int rate;
		private long s2;

		public static long getTime() {
			return System.currentTimeMillis();
		}

		public Ticker(int tickrateMS) {
			rate = tickrateMS;
			s2 = Ticker.getTime();
		}

		public int getTicks() {
			long i = Ticker.getTime();
			if (i - s2 > rate) {
				int ticks = (int) ((i - s2) / (long) rate);
				s2 += (long) rate * ticks;
				return ticks;
			}
			return 0;
		}
	}

}
